---------------------------------------------------------------------------------
-- Define Functions
---------------------------------------------------------------------------------


local smoothness_factor = 180
local laser_distance = 150 -- Distance to keep from the target

---------------------------------------------------------------------------------
-- General Setup
---------------------------------------------------------------------------------
local times_executed = ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted")
local entity = GetUpdatedEntityID()
local ex, ey = EntityGetTransform(entity)

SetRandomSeed( GameGetFrameNum(), ex + ey + entity )

local parent = EntityGetParent(entity)
local vel_x, vel_y = 0,0
local velcomp = EntityGetFirstComponent(entity, "VelocityComponent")
if ( velcomp ~= nil ) and ( velcomp > 0 ) then
	vel_x, vel_y = GameGetVelocityCompVelocity(parent)
end
local vel_mag = math.sqrt((vel_x)^2 + (vel_y)^2)
local futurex, futurey = ex + (vel_x * 60), ey + (vel_y * 60)

local lengthcomp = EntityGetFirstComponent(entity, "VariableStorageComponent", "length")
local target = {GetValueNumber("targ_x", ex), GetValueNumber("targ_y", ey)}
local length = ComponentGetValue2(lengthcomp, "value_int")
local time = GetValueNumber("time", 0)

---------------------------------------------------------------------------------
-- Inverse Kinematics
---------------------------------------------------------------------------------
-- Organize the segments and joints into neat little tables:
local legments = {}
local joints = {}
local knees = {}
for i, child in ipairs(EntityGetAllChildren(entity)) do
    if not (EntityGetName(child) == "knee") then
        local x, y = EntityGetTransform(child)
        if x == 0 then x = 1 end
        if y == 0 then y = 1 end
        table.insert(legments, child)
        if i == 1 then table.insert(joints, {ex, ey}) else table.insert(joints, {x, y}) end
    else
        table.insert(knees, child)
    end
end
table.insert(joints, target)

local luacomp = EntityGetFirstComponentIncludingDisabled(legments[2], "LuaComponent")
local laser_enabled = ComponentGetIsEnabled(luacomp)

function distance(x1, y1, x2, y2)
    return math.sqrt((x1 - x2)^2 + (y1 - y2)^2)
end

function approx_rolling_average(avg, target, n)
	if laser_enabled == true then
		n = 99999
	end
    avg = avg - avg / n
    avg = avg + target / n
    return avg
end

function move_angle_towards(current_angle, target_angle, speed)
	if laser_enabled == true then
		speed = 0.000001
	end
    if target_angle < 0 then
        target_angle = target_angle + math.pi * 2
    end
    if current_angle < 0 then
        current_angle = current_angle + math.pi * 2
    end

    local dist_up = math.abs(target_angle - current_angle)
    local dist_down = math.abs((current_angle + math.pi * 2) - target_angle)
    if current_angle > target_angle then
        dist_up = math.abs((target_angle + math.pi * 2) - current_angle)
        dist_down = math.abs(current_angle - target_angle)
    end

    local dir
    if math.abs(dist_up) < math.abs(dist_down) then
        dir = 1
    else
        dir = -1
    end
    local dist_smallest = dist_up
    if dist_down < dist_up then
        dist_smallest = dist_down
    end

    local new_angle = current_angle + speed * dir
    if dir > 0 and current_angle < target_angle and new_angle > target_angle then
        new_angle = target_angle
    elseif dir < 0 and current_angle > target_angle and new_angle < target_angle then
        new_angle = target_angle
    end
    new_angle = new_angle % (math.pi * 2)

    return new_angle
end

-- Laser arm target
local laser_target_x = GetValueNumber("laser_targ_x", futurex)
local laser_target_y = GetValueNumber("laser_targ_y", futurey)

local laser_targ_x2 = GetValueNumber("laser_targ_x2", 0)
local laser_targ_y2 = GetValueNumber("laser_targ_y2", 0)

-- local players = EntityGetWithTag("player_unit")
-- if #players > 0 then
    -- px, py = EntityGetTransform(players[1])
    -- laser_target_x = px
    -- laser_target_y = py
-- end

-- Calculate the offset for the laser target


-- Averaging for smooth movements
local function average_transform(x1, y1, x2, y2, factor)
	if laser_enabled == true then
		factor = 9999999
	end
    return approx_rolling_average(x1, x2, factor), approx_rolling_average(y1, y2, factor)
end

-- First pass, starting at end of chain:
for i = #joints, 1, -1 do
    if i == #joints then
        -- Move joint directly to the target coords:
        joints[i] = target
    elseif i ~= 1 then
        -- Move joint within range of the previous one iterated:
        local jx, jy = joints[i][1], joints[i][2]
        local jx2, jy2 = joints[i + 1][1], joints[i + 1][2]
        if jx == jx2 or jy == jy2 then jx, jy = jx + 1, jy + 1 end
        local joint_dist = distance(jx, jy, jx2, jy2)
        jx, jy = average_transform(jx2 + (((jx - jx2) / joint_dist) * length), jy2 + (((jy - jy2) / joint_dist) * length), jx, jy, smoothness_factor)
        joints[i] = {jx, jy}
    end
end

-- Second pass, starting at beginning of chain:
for i, joint in ipairs(joints) do
    if i ~= 1 then
        -- Move joint within range of the previous one iterated:
        local jx, jy = joints[i][1], joints[i][2]
        local jx2, jy2 = joints[i - 1][1], joints[i - 1][2]
        if jx == jx2 or jy == jy2 then jx, jy = jx + 1, jy + 1 end
        local joint_dist = distance(jx, jy, jx2, jy2)
        jx, jy = average_transform(jx2 + (((jx - jx2) / joint_dist) * length), jy2 + (((jy - jy2) / joint_dist) * length), jx, jy, smoothness_factor)
        joints[i] = {jx, jy}
    end
end


-- Calculate rotation towards the laser target for the second segment
local second_segment_index = 2
if #legments >= second_segment_index then
    local jx, jy = joints[second_segment_index][1], joints[second_segment_index][2]
    local vec_x, vec_y = laser_target_x - jx, laser_target_y - jy
    local targ_rot = math.atan2(vec_y, vec_x)
    local current_rot = select(3, EntityGetTransform(legments[second_segment_index]))
    targ_rot = move_angle_towards(current_rot, targ_rot, 1 / smoothness_factor)
    EntitySetTransform(legments[second_segment_index], jx, jy, targ_rot)
end

-- Arrange all the legments to connect in between the joints:
for i, legment in ipairs(legments) do
    if i ~= second_segment_index then
        local jx, jy = joints[i][1], joints[i][2]
        local jx2, jy2 = joints[i + 1][1], joints[i + 1][2]
        if jx == jx2 or jy == jy2 then jx, jy = jx + 1, jy + 1 end
        local vec_x, vec_y = jx2 - jx, jy2 - jy
        local targ_rot = math.atan2(vec_y, vec_x)
        
        -- Apply smoothing
        local avg_jx, avg_jy = average_transform(jx, jy, joints[i][1], joints[i][2], smoothness_factor)
        EntitySetTransform(legment, avg_jx, avg_jy, targ_rot)
        
        if knees[i] then EntitySetTransform(knees[i], jx2, jy2) end
    end
end

---------------------------------------------------------------------------------
-- AI:
---------------------------------------------------------------------------------
local total_leglength = length * #legments --combined length of all the segments in the leg
local length_tolerance = total_leglength --how far to let the end of the leg get before re-positioning it

--what the leg needs to point at:
local tx = GetValueNumber("tx", target[1])
local ty = GetValueNumber("ty", target[2])
local distance_to_target = distance(tx, ty, target[1], target[2])

--mix target position between ex, ey and tx, ty, based on time (picks up feet slightly when stepping):
local mult = smoothness_factor * (1 + (2 / 3))
local inv = mult - time
local mix_x = (ex / mult) * time + (tx / mult) * inv
local mix_y = (ey / mult) * time + (ty / mult) * inv

--move feet toward target, unless already within range of it, in which case snap to the coordinates:
if distance_to_target > 10 then
    target[1] = approx_rolling_average(target[1], mix_x, smoothness_factor)
    target[2] = approx_rolling_average(target[2], mix_y, smoothness_factor)
elseif target[1] ~= tx and target[2] ~= ty then
    target[1] = tx
    target[2] = ty
    --[a stepping noise could go here]
end

--if the leg stretches too far, find a new target:
if (distance(tx, ty, ex, ey) > total_leglength and distance(tx, ty, futurex, futurey) > total_leglength) or distance(tx, ty, ex, ey) == 0 then
    local surfaces_found = false
    --try to find nearby surfaces up to 3 times before resorting to stepping on air:
    for i = 1, 3 do
        potential_x, potential_y = (vel_x * 60) + math.random(-total_leglength, total_leglength), (vel_y * 60) + math.random(-total_leglength, total_leglength)
        mag = math.sqrt(((ex + potential_x) - ex)^2 + ((ey + potential_y) - ey)^2)
        potential_x, potential_y = ex + ((potential_x / mag) * total_leglength), ey + ((potential_y / mag) * total_leglength)
        did_hit, hit_x, hit_y = RaytracePlatforms(ex, ey, potential_x, potential_y)
        if did_hit then
            tx = hit_x
            ty = hit_y
            surfaces_found = true
            break
        end
    end
    --step on air:
    if not surfaces_found then
        tx = potential_x
        ty = potential_y
    end
    time = mult
end

--apply changes:
SetValueNumber("tx", tx)
SetValueNumber("ty", ty)
SetValueNumber("time", math.max(0, time - 1))
SetValueNumber("targ_x", approx_rolling_average(target[1], laser_target_x, smoothness_factor))
SetValueNumber("targ_y", approx_rolling_average(target[2], laser_target_y, smoothness_factor))

local testx, testy = DEBUG_GetMouseWorld()

if (times_executed == 1) or (times_executed % 600 == 0) then
    -- Set a new target and disable the laser
    
	SetValueNumber("laser_targ_x2", testx)
	SetValueNumber("laser_targ_y2", testy)
	
    EntitySetComponentsWithTagEnabled(legments[2], "enabled_in_world", false)
    -- GamePrint("Laser disabled and new target set")
elseif (times_executed % 600 == 300) then
    -- Enable the laser without updating the target
    EntitySetComponentsWithTagEnabled(legments[2], "enabled_in_world", true)
	ComponentSetValue2(luacomp, "mTimesExecuted", 0)
	GameEntityPlaySound( legments[2], "oruai_buildup" )
	
    -- GamePrint("Laser enabled without updating target")
else
    -- Prevent target updates while the laser is enabled

        -- Perform regular movement calculations
        SetValueNumber("targ_x", approx_rolling_average(target[1], laser_target_x, smoothness_factor))
        SetValueNumber("targ_y", approx_rolling_average(target[2], laser_target_y, smoothness_factor))
    -- end
end


SetValueNumber("laser_targ_x", approx_rolling_average(testx, testx + laser_targ_x2, 1))
SetValueNumber("laser_targ_y", approx_rolling_average(testy, testy + laser_targ_y2, 1))