function approx_rolling_average(avg, target, n)
    avg = avg - avg / n;
    avg = avg + target / n;
    return avg
end
local smoothness_factor = 20

local function get_state()
	local entity_id = GetUpdatedEntityID()
	_state = _state or {}
	_state[entity_id] = _state[entity_id] or {}
	for i, v in ipairs(_state) do
		if not EntityGetIsAlive(v) then
			-- Remove saved states of entities that no longer exist
			-- to prevent memory leak
			_state[entity_id] = nil
		end
	end
	return _state[entity_id]
end

local function distance(x1, y1, x2, y2)
  local dx = x2 - x1
  local dy = y2 - y1
  return math.sqrt(dx * dx + dy * dy)
end
-- Function to subdivide a 2D segmented line into parts of equal length
local function subdivide_line(segmented_line, equal_length)
  if #segmented_line < 2 or equal_length <= 0 then
      return nil
  end

  local total_length = 0
  for i = 1, #segmented_line - 1 do
      local x1, y1 = segmented_line[i].x, segmented_line[i].y
      local x2, y2 = segmented_line[i + 1].x, segmented_line[i + 1].y
      total_length = total_length + distance(x1, y1, x2, y2)
  end

  local num_subdivisions = math.floor(total_length / equal_length)
  if num_subdivisions == 0 then
      return nil
  end

  local segment_length = total_length / num_subdivisions

  local result = {}
  table.insert(result, segmented_line[1]) -- Include the starting point

  local current_length = 0
  for i = 1, #segmented_line - 1 do
      local x1, y1 = segmented_line[i].x, segmented_line[i].y
      local x2, y2 = segmented_line[i + 1].x, segmented_line[i + 1].y
      local segment_dist = distance(x1, y1, x2, y2)

      while current_length + segment_dist >= segment_length do
          local t = (segment_length - current_length) / segment_dist
          local new_x = x1 + t * (x2 - x1)
          local new_y = y1 + t * (y2 - y1)
          table.insert(result, { x = new_x, y = new_y })
          segment_dist = segment_dist - (segment_length - current_length)
          x1, y1 = new_x, new_y
          current_length = 0
      end

      current_length = current_length + segment_dist
  end

  -- Include the ending point
  table.insert(result, segmented_line[#segmented_line])

  return result
end

local function map(value, old_min, old_max, new_min, new_max)
	return (value - old_min) * (new_max - new_min) / (old_max - old_min) + new_min;
end

-- Function to calculate the arc points with adjustable amplitude
local function calculate_arc(start_x, start_y, end_x, end_y, num_points, concave, amplitude)
    local points = {}
    local mid_x = (start_x + end_x) / 2
    local mid_y = (start_y + end_y) / 2

    local control_x = mid_x
    local control_y = mid_y + (concave and -amplitude or amplitude)  -- Use amplitude to control the concavity

    for t = 0, 1, 1 / (num_points - 1) do
        local u = 1 - t
        local tt = t * t
        local uu = u * u

        local x = uu * start_x + 2 * u * t * control_x + tt * end_x
        local y = uu * start_y + 2 * u * t * control_y + tt * end_y

        table.insert(points, { x = x, y = y })
    end

    return points
end

-- Function to draw limb using sprite components
local function draw_limb(entity_id, start_x, start_y, end_x, end_y, scale_x, amplitude)
    local num_points = 10  -- Adjust the number of points for smoothness
    local points = calculate_arc(start_x, start_y, end_x, end_y, num_points, scale_x > 0, amplitude)
    local half_point = #points / 2
    local sprite_size = 9
    local max_length = 100

    -- Calculate total length of the arc
    local path_length = 0
    for i = 1, #points - 1 do
        path_length = path_length + distance(points[i].x, points[i].y, points[i + 1].x, points[i + 1].y)
    end

    -- Adjust scaling
    local dist_p = math.min(1, path_length / max_length)
    dist_p = map(dist_p, 0, 1, 3.5, 2)
    local new_path = subdivide_line(points, dist_p) or {}

    -- Create or remove SpriteComponents until it matches the amount of points we have on the path
    local sprite_comps = EntityGetComponentIncludingDisabled(entity_id, "SpriteComponent") or {}
    -- Create new comps until we have enough
    for i = 1, #new_path - #sprite_comps do
        EntityAddComponent2(entity_id, "SpriteComponent", {
            image_file = "data/enemies_gfx/manticore_segment.png",
            update_transform_rotation = false,
            has_special_scale = true,
            z_index = 8.1
        })
    end
    -- Remove comps that are too many
    for i = 1, #sprite_comps - #new_path do
        EntityRemoveComponent(entity_id, sprite_comps[i])
    end
    sprite_comps = EntityGetComponentIncludingDisabled(entity_id, "SpriteComponent") or {}

    -- Set SpriteComponent offsets
    for i, v in ipairs(new_path) do
        local h = (i - half_point) / #new_path
        h = h * h
        local scale = 1 - math.min(max_length, path_length) / max_length
        scale = scale + h * 2
        scale = math.min(1, math.max(0.8, scale))

        ComponentSetValue2(sprite_comps[i], "offset_x", -(v.x - start_x) * (1 / scale) + (sprite_size / 2))
        ComponentSetValue2(sprite_comps[i], "offset_y", -(v.y - start_y) * (1 / scale) + (sprite_size / 2))
        ComponentSetValue2(sprite_comps[i], "special_scale_x", scale)
        ComponentSetValue2(sprite_comps[i], "special_scale_y", scale)
    end
end

-- Integrate the draw_limb function with the IK limb script
local entity = GetUpdatedEntityID()
local ex, ey, r, sx = EntityGetTransform(entity)

local parent = EntityGetRootEntity(entity)
local parent_x, parent_y, pr, psx = EntityGetTransform(parent)
local target = {GetValueNumber("targ_x", ex), GetValueNumber("targ_y", ey)}

-- Use the draw_limb function for each limb segment with reduced amplitude
local amplitude = 25  -- Adjust this value to control the arc height


local cx, cy = GameGetCameraPos()
if (ex > cx + 256) or
	(ex < cx - 256) or
	(ey < cy - 150) or
	(ey > cy + 150) then
	draw_limb(entity, ex, ey, ex, ey + 20, -1, amplitude)
	ComponentSetValue2(GetUpdatedComponentID(), "execute_every_n_frame", 80)
else
	draw_limb(entity, ex, ey, target[1], target[2] + 20, -1, amplitude)
	ComponentSetValue2(GetUpdatedComponentID(), "execute_every_n_frame", 1)
end


local ik_limb_comp = EntityGetFirstComponentIncludingDisabled(entity, "IKLimbWalkerComponent")
local new_x, new_y = ComponentGetValue2(ik_limb_comp, "mTarget")

if new_y < ey then
	new_y = ey + (ey - new_y)
end

local average_x = approx_rolling_average(target[1], new_x, smoothness_factor)
local average_y = approx_rolling_average(target[2], new_y, smoothness_factor)

SetValueNumber("targ_x", average_x)
SetValueNumber("targ_y", average_y)
